"""Serial training of AlexNet on CIFAR-10."""
import torch

from utils_data import get_dataloaders_cifar10, get_transforms_cifar10
from utils_train import train_model
from model import AlexNet

if __name__ == "__main__":
    # Set random seed.
    seed = 123
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    b = 256  # Set batch size.
    e = 100  # Set number of epochs to be trained.
    data_root = "/pfs/work7/workspace/scratch/ku4408-VL-ScalableAI/data/cifar"

    device = torch.device(
        "cuda:0" if torch.cuda.is_available() else "cpu"
    )  # Check device.
    print(f"Using {device} device.")

    # DATASET
    # Using transforms on your data allows you to take it from its source state
    # and transform it into data that’s ready for training.

    # Get transforms applied to CIFAR-10 data for training and inference.
    train_transforms, test_transforms = get_transforms_cifar10()

    # Get PyTorch dataloaders for training, testing, and validation dataset.
    train_loader, valid_loader, test_loader = get_dataloaders_cifar10(
        batch_size=b,  # The batch size.
        data_root=data_root,  # The path to the data dir.
        validation_fraction=0.1,  # The validation fraction.
        train_transforms=train_transforms,  # The transforms applied to the data at training time.
        test_transforms=test_transforms,  # The transforms applied to the data at inference time.
    )

    # Check loaded dataset.
    for images, labels in train_loader:
        print("Image batch dimensions:", images.shape)
        print("Image label dimensions:", labels.shape)
        print("Class labels of 10 examples:", labels[:10])
        break

    # MODEL
    # Define neural network by subclassing PyTorch's nn.Module.

    model = AlexNet(num_classes=10).to(
        device
    )  # Build instance of AlexNet with 10 classes and move it to device.

    # torch.optim package implements various optimization algorithms.
    # To use torch.optim you have to construct an optimizer object,
    # that will hold the current state and will update the parameters
    # based on computed gradients.
    # Stochastic gradient descent (optionally with momentum):
    optimizer = torch.optim.SGD(model.parameters(), momentum=0.9, lr=0.1)

    # torch.optim.lr_scheduler provides several learning-rate adjustment methods based on number of epochs.
    # torch.optim.lr_scheduler.ReduceLROnPlateau: dynamic learning rate reducing based on some validation measurements.
    # Reduce learning rate when a metric has stopped improving:
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer, factor=0.1, mode="max",
    )

    # Train model.
    loss_history, train_acc_history, valid_acc_history = train_model(
        model=model,  # The model to train.
        num_epochs=e,  # The number of epochs to train.
        train_loader=train_loader,  # The training dataloader.
        valid_loader=valid_loader,  # The validation dataloader.
        test_loader=test_loader,  # The test dataloader.
        optimizer=optimizer,  # The optimizer.
        device=device,  # The device to train on.
        scheduler=scheduler,  # The learning rate scheduler.
        logging_interval=100,  # The logging interval.
    )

    # Save loss and accuracy metrics to disk.
    torch.save(loss_history, "loss.pt")
    torch.save(train_acc_history, "train_acc.pt")
    torch.save(valid_acc_history, "valid_acc.pt")
