"""Distributed data-parallel training of AlexNet on CIFAR-10."""
import os

import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as Ddp
import torchvision

from utils_data import get_dataloaders_cifar10_ddp, get_transforms_cifar10
from utils_train import train_model_ddp
from utils_eval import compute_accuracy_ddp
from model import AlexNet


def main():
    """Distributed data-parallel training of AlexNet on CIFAR-10."""
    world_size = int(os.getenv("SLURM_NPROCS"))  # Get overall number of processes.
    rank = int(os.getenv("SLURM_PROCID"))  # Get individual process ID.
    data_root = "/pfs/work7/workspace/scratch/ku4408-VL-ScalableAI/data"
    print(
        f"Rank, world size, device count: {rank}, {world_size}, {torch.cuda.device_count()}"
    )

    if rank == 0:
        if dist.is_available():
            print(
                "Distributed package available...[OK]"
            )  # Check if distributed package available.
        if dist.is_nccl_available():
            print("NCCL backend available...[OK]")  # Check if NCCL backend available.

    # On each host with N GPUs, spawn up N processes, while ensuring that
    # each process individually works on a single GPU from 0 to N-1.

    address = os.getenv("SLURM_LAUNCH_NODE_IPADDR")
    port = "29500"
    os.environ["MASTER_ADDR"] = address
    os.environ["MASTER_PORT"] = port

    # Initialize DDP.
    dist.init_process_group(backend="nccl", rank=rank, world_size=world_size)
    torch.cuda.set_device(rank % torch.cuda.device_count())
    if dist.is_initialized():
        print("Process group initialized successfully...[OK]")  # Check initialization.

    # Check used backend.
    print(dist.get_backend(), "backend used...[OK]")

    b = 256  # Set batch size.
    e = 100  # Set number of epochs to be trained.

    # DATASET
    # Using transforms on your data allows you to take it from its source state
    # and transform it into data that’s ready for training.

    # Get transforms applied to CIFAR-10 data for training and inference.
    train_transforms, test_transforms = get_transforms_cifar10()

    # Get distributed dataloaders for training and validation data on all ranks.
    train_loader, valid_loader = get_dataloaders_cifar10_ddp(
        batch_size=b,
        data_root=data_root,
        train_transforms=train_transforms,
        test_transforms=test_transforms,
    )

    model = AlexNet(
        num_classes=10
    ).cuda()  # Create model and move it to GPU with id rank.
    ddp_model = Ddp(model)  # Wrap model with DDP.
    optimizer = torch.optim.SGD(
        ddp_model.parameters(), momentum=0.9, lr=0.1
    )  # Set up optimizer and pass DDP parameters.

    # Train model.
    train_model_ddp(
        model=ddp_model,
        num_epochs=e,
        train_loader=train_loader,
        valid_loader=valid_loader,
        optimizer=optimizer,
    )

    # Test final model on root.
    if dist.get_rank() == 0:
        test_dataset = torchvision.datasets.CIFAR10(
            root=data_root,
            train=False,
            transform=test_transforms,
        )  # Load test dataset.
        test_loader = torch.utils.data.DataLoader(
            dataset=test_dataset, batch_size=b, shuffle=False
        )  # Set up test dataloader.
        test_acc = compute_accuracy_ddp(
            ddp_model, test_loader
        )  # Compute accuracy on test data.
        print(f"Test accuracy {test_acc :.2f}%")

    dist.destroy_process_group()  # Destroy process group.


if __name__ == "__main__":
    main()
