"""Helper functions for training (DDP) AlexNet on CIFAR-10."""
import time
from typing import List, Tuple, Union

import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau

from utils_eval import compute_accuracy, get_right_ddp


def train_model(
    model: torch.nn.Module,
    num_epochs: int,
    train_loader: torch.utils.data.DataLoader,
    valid_loader: torch.utils.data.DataLoader,
    test_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    device: torch.device,
    logging_interval: int = 50,
    scheduler: Union[torch.optim.lr_scheduler._LRScheduler, ReduceLROnPlateau] = None,
) -> Tuple[List[float], List[float], List[float]]:
    """
    Train your model.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train.
    num_epochs : int
        The number of epochs to train
    train_loader : torch.utils.data.DataLoader
        The training dataloader.
    valid_loader : torch.utils.data.DataLoader
        The validation dataloader.,
    test_loader : torch.utils.data.DataLoader
        The testing dataloader.
    optimizer : torch.optim.Optimizer
        The optimizer to use.
    device : torch.device
        The device to train on.
    logging_interval : int
        The logging interval.
    scheduler : torch.optim.lr_scheduler._LRScheduler
        An optional learning rate scheduler.

    Returns
    -------
    List[float]
        The loss history.
    List[float]
        The training accuracy history.
    List[float]
        The validation accuracy history.
    """
    start = time.perf_counter()  # Measure training time.

    # Initialize history lists for loss, training accuracy, and validation accuracy.
    loss_history, train_acc_history, valid_acc_history = [], [], []

    for epoch in range(num_epochs):  # Loop over epochs.
        model.train()  # Set model to training mode.
        # Thus, layers like dropout which behave differently on train and test procedures
        # know what is going on and can behave accordingly. model.train() sets the mode to
        # train. One might expect this to train model but it does not do that.
        # Call either model.eval() or model.train(mode=False) to tell that you are testing.

        for batch_idx, (features, targets) in enumerate(
            train_loader
        ):  # Loop over mini batches.
            features = features.to(device)  # Move features to used device.
            targets = targets.to(device)  # Move targets to used device.

            # Forward and backward pass.
            logits = model(features)  # Calculate logits as the model's output.
            loss = torch.nn.functional.cross_entropy(
                logits, targets
            )  # Calculate cross-entropy loss.
            optimizer.zero_grad()  # Zero out gradients.
            # Gradients are accumulated and not overwritten whenever .backward() is called.
            loss.backward()  # Calculate gradients of loss w.r.t. model parameters in backward pass.
            optimizer.step()  # Perform single optimization step to update model parameters.

            # Logging.
            loss_history.append(loss.item())
            if not batch_idx % logging_interval:
                print(
                    f"Epoch: {epoch+1:03d}/{num_epochs:03d} "
                    f"| Batch {batch_idx:04d}/{len(train_loader):04d} "
                    f"| Loss: {loss:.4f}"
                )

        model.eval()  # Set model to evaluation mode.

        with torch.no_grad():  # Disable gradient calculation to reduce memory consumption.
            train_acc = compute_accuracy(
                model, train_loader, device=device
            )  # Compute accuracy on training data.
            valid_acc = compute_accuracy(
                model, valid_loader, device=device
            )  # Compute accuracy on validation data.
            print(
                f"Epoch: {epoch+1:03d}/{num_epochs:03d} "
                f"| Train: {train_acc :.2f}% "
                f"| Validation: {valid_acc :.2f}%"
            )
            train_acc_history.append(train_acc)
            valid_acc_history.append(valid_acc)

        elapsed = (time.perf_counter() - start) / 60  # Measure training time per epoch.
        print(f"Time elapsed: {elapsed:.2f} min")

        if scheduler is not None:
            original_lr = scheduler.get_last_lr()[0]
            scheduler.step(valid_acc_history[-1])
            new_lr = scheduler.get_last_lr()[0]
            if original_lr != new_lr:
                print(f"Epoch: {epoch+1:03d}/{num_epochs:03d}: LR updated to {new_lr}")

    elapsed = (time.perf_counter() - start) / 60  # Measure total training time.
    print(f"Total Training Time: {elapsed:.2f} min")

    test_acc = compute_accuracy(
        model, test_loader, device=device
    )  # Compute accuracy on test data.
    print(f"Test accuracy {test_acc :.2f}%")

    return loss_history, train_acc_history, valid_acc_history


def train_model_ddp(
    model: torch.nn.Module,
    num_epochs: int,
    train_loader: torch.utils.data.DataLoader,
    valid_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
) -> Tuple[List[float], List[float], List[float]]:
    """
    Train the model in distributed data-parallel fashion.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train.
    num_epochs : int
        The number of epochs to train.
    train_loader : torch.utils.data.DataLoader
        The training dataloader.
    valid_loader : torch.utils.data.DataLoader
        The validation dataloader.
    optimizer : torch.optim.Optimizer
        The optimizer to use.

    Returns
    -------
    List[float]
        The epoch-wise loss history.
    List[float]
        The epoch-wise training accuracy history.
    List[float]
        The epoch-wise validation accuracy history.
    """
    start = time.perf_counter()  # Measure training time.
    rank = torch.distributed.get_rank()
    world_size = torch.distributed.get_world_size()

    loss_history, train_acc_history, valid_acc_history = (
        [],
        [],
        [],
    )  # Initialize history lists.

    for epoch in range(num_epochs):  # Loop over epochs.
        train_loader.sampler.set_epoch(epoch)
        model.train()  # Set model to training mode.

        for batch_idx, (features, targets) in enumerate(
            train_loader
        ):  # Loop over mini batches.
            features = features.cuda()
            targets = targets.cuda()

            # Forward and backward pass.
            logits = model(features)
            loss = torch.nn.functional.cross_entropy(logits, targets)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()  # Perform single optimization step to update model parameters.

            # Logging.
            torch.distributed.all_reduce(
                loss
            )  # Allreduce rank-local mini-batch losses.
            loss /= world_size  # Average all-reduced rank-local mini-batch losses over all ranks.
            loss_history.append(
                loss.item()
            )  # Append globally averaged loss of this epoch to history list.

            if rank == 0:
                print(
                    f"Epoch: {epoch+1:03d}/{num_epochs:03d} "
                    f"| Batch {batch_idx:04d}/{len(train_loader):04d} "
                    f"| Averaged Loss: {loss:.4f}"
                )

        model.eval()  # Set model to evaluation mode.

        with torch.no_grad():  # Disable gradient calculation.
            # Get rank-local numbers of correctly classified and overall samples in training and validation set.
            right_train, num_train = get_right_ddp(model, train_loader)
            right_valid, num_valid = get_right_ddp(model, valid_loader)

            torch.distributed.all_reduce(right_train)
            torch.distributed.all_reduce(right_valid)
            torch.distributed.all_reduce(num_train)
            torch.distributed.all_reduce(num_valid)
            train_acc = right_train.item() / num_train.item() * 100
            valid_acc = right_valid.item() / num_valid.item() * 100
            train_acc_history.append(train_acc)
            valid_acc_history.append(valid_acc)

            if rank == 0:
                print(
                    f"Epoch: {epoch+1:03d}/{num_epochs:03d} "
                    f"| Train: {train_acc :.2f}% "
                    f"| Validation: {valid_acc :.2f}%"
                )

        elapsed = (time.perf_counter() - start) / 60  # Measure training time per epoch.
        elapsed = torch.Tensor([elapsed]).cuda()
        torch.distributed.all_reduce(elapsed)
        elapsed /= world_size
        if rank == 0:
            print(f"Time elapsed: {elapsed.item()} min")

    elapsed = (time.perf_counter() - start) / 60  # Measure total training time.
    elapsed = torch.Tensor([elapsed]).cuda()
    torch.distributed.all_reduce(elapsed)
    elapsed /= world_size

    if rank == 0:
        print("Total Training Time:", elapsed.item(), "min")
        torch.save(loss_history, f"loss_{world_size}_gpu.pt")
        torch.save(train_acc_history, f"train_acc_{world_size}_gpu.pt")
        torch.save(valid_acc_history, f"valid_acc_{world_size}_gpu.pt")

    return loss_history, train_acc_history, valid_acc_history
