"""Functions to optimize."""
# Define functions to optimize, in this case minimize.
# Taken from https://towardsdatascience.com/
# optimization-eye-pleasure-78-benchmark-test-functions-for-
# single-objective-optimization-92e7ed1d1f12
# Save to separate module file `functions.py`.

from typing import Callable, Dict, Tuple

import numpy as np


def bukin_n6(params: Dict[str, float]) -> float:
    """
    Bukin N.6 function: continuous, convex, non-separable, non-differentiable, multimodal

    Input domain: -15 <= x <= -5, -3 <= y <= 3
    Global minimum 0 at (x, y) = (-10, 1)

    Parameters
    ----------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    params = np.array(list(params.values()))
    return 100 * np.sqrt(np.abs(params[1] - 0.01 * params[0] ** 2)) + 0.01 * np.abs(
        params[0] + 10
    )


def egg_crate(params: Dict[str, float]) -> float:
    """
    Egg-crate function: continuous, non-convex, separable, differentiable, multimodal

    Input domain: -5 <= x, y <= 5
    Global minimum -1 at (x, y) = (0, 0)

    Parameters
    ----------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    params = np.array(list(params.values()))
    return (
        params[0] ** 2
        + params[1] ** 2
        + 25 * (np.sin(params[0]) ** 2 + np.sin(params[1]) ** 2)
    )


def himmelblau(params: Dict[str, float]) -> float:
    """
    Himmelblau function: continuous, non-convex, non-separable, differentiable, multimodal

    Input domain: -6 <= x, y <= 6
    Global minimum 0 at (x, y) = (3, 2)

    Parameters
    ----------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    params = np.array(list(params.values()))
    return (params[0] ** 2 + params[1] - 11) ** 2 + (
        params[0] + params[1] ** 2 - 7
    ) ** 2


def keane(params: Dict[str, float]) -> float:
    """
    Keane function: continuous, non-convex, non-separable, differentiable, multimodal

    Input domain: -10 <= x, y <= 10
    Global minimum 0.6736675 at (x, y) = (1.3932491, 0) and (x, y) = (0, 1.3932491)

    Parameters
    ------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    params = np.array(list(params.values()))
    return (
        -np.sin(params[0] - params[1]) ** 2
        * np.sin(params[0] + params[1]) ** 2
        / np.sqrt(params[0] ** 2 + params[1] ** 2)
    )


def leon(params: Dict[str, float]) -> float:
    """
    Leon function: continuous, non-convex, non-separable, differentiable, non-multimodal, non-random, non-parametric

    Input domain: 0 <= x, y <= 10
    Global minimum 0 at (x, y) =(1, 1)

    Parameters
    ----------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    params = np.array(list(params.values()))
    return 100 * (params[1] - params[0] ** 3) ** 2 + (1 - params[0]) ** 2


def rastrigin(params: Dict[str, float]) -> float:
    """
    Rastrigin function: continuous, non-convex, separable, differentiable, multimodal

    A non-linear and highly multimodal function. Its surface is determined by two external variables, controlling
    the modulation’s amplitude and frequency. The local minima are located at a rectangular grid with size 1.
    Their functional values increase with the distance to the global minimum.

    Input domain: -5.12 <= x_i <= 5.12, i = 1,...,N
    Global minimum 0 at (x_i)_N = (0)_N

    Parameters
    ----------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    a = 10.0
    params = np.array(list(params.values()))
    return a * len(params) + np.sum(params**2 - a * np.cos(2 * np.pi * params))


def schwefel(params: Dict[str, float]) -> float:
    """
    Schwefel 2.20 function: continuous, convex, separable, non-differentiable, non-multimodal

    This function has a second-best minimum far away from the global optimum.

    Input domain: -500 <= x_i <= 500, i = 1,...,N
    Global minimum 0 at (x_i)_N = (420.968746)_N

    Parameters
    ----------
    params : dict[str, float]
        The function parameters

    Returns
    -------
    float
        The function value.
    """
    v = 418.982887
    params = np.array(list(params.values()))
    return v * len(params) - np.sum(params * np.sin(np.sqrt(np.abs(params))))


def sphere(params: Dict[str, float]) -> float:
    """
    Sphere function: continuous, convex, separable, differentiable, unimodal

    Input domain: -5.12 <= x, y <= 5.12
    Global minimum 0 at (x, y) = (0, 0)

    Parameters
    ----------
    params : dict[str, float]
        The function parameters.

    Returns
    -------
    float
        The function value.
    """
    return np.sum(np.array(list(params.values())) ** 2)


def get_function_search_space(
    fname: str,
) -> Tuple[Callable, Dict[str, Tuple[float, float]]]:
    """
    Get search space limits and function from function name.

    Parameters
    ----------
    fname : str
        The function name.

    Returns
    -------
    Callable
        The function.
    dict[str, tuple[float, float]]
        The function's search space.
    """
    if fname == "bukin":
        function = bukin_n6
        limits = {
            "x": [-15.0, -5.0],
            "y": [-3.0, 3.0],
        }
    elif fname == "eggcrate":
        function = egg_crate
        limits = {
            "x": [-5.0, 5.0],
            "y": [-5.0, 5.0],
        }
    elif fname == "himmelblau":
        function = himmelblau
        limits = {
            "x": [-6.0, 6.0],
            "y": [-6.0, 6.0],
        }
    elif fname == "keane":
        function = keane
        limits = {
            "x": [-10.0, 10.0],
            "y": [-10.0, 10.0],
        }
    elif fname == "leon":
        function = leon
        limits = {
            "x": [0.0, 10.0],
            "y": [0.0, 10.0],
        }

    elif fname == "rastrigin":
        function = rastrigin
        limits = {
            "x": [-5.12, 5.12],
            "y": [-5.12, 5.12],
        }

    elif fname == "schwefel":
        function = schwefel
        limits = {
            "x": [-500.0, 500.0],
            "y": [-500.0, 500.0],
        }

    elif fname == "sphere":
        function = sphere
        limits = {
            "x": [-5.12, 5.12],
            "y": [-5.12, 5.12],
        }

    else:
        ValueError(f"Function {fname} undefined...exiting")

    return function, limits
