"""Serial optimization of mathematical functions with Hyppopy."""
from hyppopy.solvers.GridsearchSolver import (
    GridsearchSolver,
)  # Import grid-search solver.
from hyppopy.solvers.OptunitySolver import (
    OptunitySolver,
)  # Import PSO solver (based on optunity library).
from hyppopy.solvers.RandomsearchSolver import (
    RandomsearchSolver,
)  # Import random-search solver.

import functions

if __name__ == "__main__":
    # -------- Setup --------
    # Set frequency for grid-search solver.
    frequency_x = 100
    frequency_y = 100

    # Set number of iterations for random-search and PSO solver.
    # Overall number of tested samples should be the same for all solvers.
    max_iterations = frequency_x * frequency_y
    # Get function and search space from function name.
    fname = "sphere"
    function, limits = functions.get_function_search_space(fname)
    config = {
        "hyperparameter": {
            "x": {
                "domain": "uniform",
                "data": limits["x"],
                "type": float,
                "frequency": frequency_x,
            },
            "y": {
                "domain": "uniform",
                "data": limits["y"],
                "type": float,
                "frequency": frequency_y,
            },
        },
        "max_iterations": max_iterations,
    }

    # Note: While the grid-search solver determines its grid and thus the number of probed samples
    # based on `frequency`, the random-search solver and the PSO solver determine the number of
    # probed samples from `max_iterations`.

    solver_gs = GridsearchSolver(config)  # Create grid search solver instance.
    solver_rs = RandomsearchSolver(config)  # Create random search solver instance.
    solver_pso = OptunitySolver(config)  # Create PSO solver instance.

    # Pass loss function to solver.
    solver_gs.blackbox = function
    solver_rs.blackbox = function
    solver_pso.blackbox = function

    # -------- Optimization --------
    print(
        "#############################################################\n"
        "# HYPPOPY: a HYPer-Parameter Optimization toolbox in PYthon #\n"
        "#############################################################\n"
        f"Tested function: {function.__name__}\n",
    )

    # Grid search
    print("###############\n# GRID SEARCH #\n###############")
    solver_gs.run()  # Run grid-search solver.
    _, best_gs = solver_gs.get_results()
    print(f"Minimum found is {solver_gs.blackbox(best_gs)}.\n")

    # Random search
    print("#################\n# RANDOM SEARCH #\n#################")
    solver_rs.run()  # Run random-search solver.
    _, best_rs = solver_rs.get_results()
    print(f"Minimum found is {solver_rs.blackbox(best_rs)}.\n")

    # PSO
    print("#######\n# PSO #\n#######")
    solver_pso.run()  # Run PSO solver.
    _, best_pso = solver_pso.get_results()
    print(f"Minimum found is {solver_pso.blackbox(best_pso)}.")
