"""MPI-parallel optimization of mathematical functions with Hyppopy."""
import argparse

from mpi4py import MPI
from hyppopy.HyppopyProject import HyppopyProject
from hyppopy.MPIBlackboxFunction import MPIBlackboxFunction
from hyppopy.SolverPool import SolverPool
from hyppopy.solvers.MPISolverWrapper import MPISolverWrapper

import functions

if __name__ == "__main__":
    comm = MPI.COMM_WORLD  # Set up MPI communicator.
    rank, size = comm.rank, comm.size  # Get individual rank of process and world size.

    if rank == 0:
        print("#############################################################")
        print("# HYPPOPY: a HYPer-Parameter Optimization toolbox in PYthon #")
        print("#############################################################")

    comm.barrier()
    print(f"On rank {rank} / {size}.")

    # Pass function name and used solver as command-line arguments.
    parser = argparse.ArgumentParser(
        prog="Distributed Optimization of Mathematical Functions with Hyppopy"
    )
    parser.add_argument(
        "--function",
        type=str,
        choices=[
            "bukin",
            "eggcrate",
            "himmelblau",
            "keane",
            "leon",
            "rastrigin",
            "schwefel",
            "sphere",
        ],
        default="sphere",
        help="The function to optimize.",
    )

    parser.add_argument(
        "--solver",
        type=str,
        choices=["gridsearch", "randomsearch"],
        default="gridsearch",
        help="The solver to use.",
    )

    args = parser.parse_args()

    # Set frequency for grid-search solver.
    frequency_x = 100
    frequency_y = 100

    # Set number of iterations for random-search solver.
    max_iterations = frequency_x * frequency_y
    # Get function and search space from function name.
    function, limits = functions.get_function_search_space(args.function)
    config = {
        "hyperparameter": {
            "x": {
                "domain": "uniform",
                "data": limits["x"],
                "type": float,
                "frequency": frequency_x,
            },
            "y": {
                "domain": "uniform",
                "data": limits["y"],
                "type": float,
                "frequency": frequency_y,
            },
        },
        "max_iterations": max_iterations,
        "solver": args.solver,
    }

    project = HyppopyProject(config=config)
    solver = MPISolverWrapper(solver=SolverPool.get(project=project), mpi_comm=comm)
    blackbox = MPIBlackboxFunction(blackbox_func=function, mpi_comm=comm)
    solver.blackbox = blackbox
    solver.run()

    df, best = solver.get_results()

    if solver.is_master() is True:
        print("\n")
        print("*" * 110)
        print(f"Considered function: {function.__name__}")
        print(f"Potential minimum {function(best)} found at {best}.")
        print("*" * 110)
