import pathlib
import pickle
from typing import List, Union

from matplotlib import pyplot as plt
import numpy as np
import propulate


def load_pop(path: Union[pathlib.Path, str]):
    """
    Load population list from propulate checkpoint.

    Parameters
    ----------
    path : str | pathlib.Path
        The path to the checkpoint file

    Returns
    -------
    list
        The population list
    """
    with open(pathlib.Path(path), "rb") as f:
        return pickle.load(f)


def get_losses_propulate(pop: List[propulate.Individual]):
    """
    Get losses and evaluation times from propulate population list.

    Parameters
    ----------
    pop : list
        The population list

    Returns
    -------
    times : np.ndarray
        The sorted evaluation times
    losses : np.ndarray
        The corresponding losses
    generations : np.ndarray
        The corresponding generations
    """
    actives = [ind for ind in pop if ind.active is True]
    actives.sort(key=lambda x: x.evaltime)
    losses = np.array([ind.loss for ind in actives])
    times = np.array([ind.evaltime for ind in actives])
    generations = np.array([ind.generation for ind in actives])
    ranks = np.array([ind.rank for ind in actives])
    times = times - times[0]

    return times, losses, generations, ranks


if __name__ == "__main__":
    path = "./torch_ckpts/island_0_ckpt.pickle"
    population = load_pop(path)
    print(population)

    times, losses, generations, ranks = get_losses_propulate(population)

    plt.figure(figsize=(5, 3))
    plt.scatter(x=times / 3600, y=-losses * 100, c=generations)
    plt.plot(
        times[np.argmax(-losses)] / 3600,
        np.amax(-losses) * 100,
        marker="*",
        ms=12,
        mfc="white",
        mec="k",
    )
    plt.xlabel("Runtime / h", fontweight="bold")
    plt.ylabel("Validation accuracy / %", fontweight="bold")
    plt.grid()
    cbar = plt.colorbar()
    cbar.ax.set_ylabel("Generation", weight="bold")
    plt.tight_layout()
    plt.savefig("val_acc_generation.pdf")
    plt.show()

    plt.figure(figsize=(5, 3))
    plt.scatter(
        x=times / 3600,
        y=-losses * 100,
        c=ranks,
        cmap="viridis",
        s=10 * (generations + np.ones_like(generations)) ** 0.75,
    )
    plt.plot(
        times[np.argmax(-losses)] / 3600,
        np.amax(-losses) * 100,
        marker="*",
        ms=12,
        mfc="white",
        mec="k",
    )
    plt.xlabel("Runtime / h", fontweight="bold")
    plt.ylabel("Validation accuracy / %", fontweight="bold")
    plt.grid()
    cbar = plt.colorbar(ticks=[0, 1, 2, 3])
    cbar.ax.set_ylabel("Worker rank", weight="bold")
    plt.tight_layout()
    plt.savefig("val_acc_rank.pdf")
    plt.show()
